@@ -6,6 +6,7 @@ import io
 import itertools
 import json
 import logging
+import ntpath
 import os
 import random
 import sys
@@ -114,13 +115,18 @@ def _gen_rand():
     zipfile.ZIP_DEFLATED,
     zipfile.ZIP_BZIP2
 ])
-def test_zipstream_compression(files, ct):
+def test_zipstream_compression(caplog, files, ct):
     """Test that all types of compression properly compress and extract"""
+    caplog.set_level(logging.WARNING)
+
     zs = ZipStream(compress_type=ct)
     for f in files:
         zs.add_path(f[0])
 
     zf = _get_zip(zs)
+
+    assert "Size mismatch" not in caplog.text
+
     zinfos = zf.infolist()
     assert len(zinfos) == len(files)
 
@@ -241,7 +247,7 @@ def test_footer_partial(files):
     zipfile.ZIP_STORED,
     zipfile.ZIP_LZMA
 ])
-def test_compress_level_python_36(monkeypatch, ct):
+def test_compress_level_python_36(ct):
     """Test that using compress_level on <3.7 produces an error"""
 
     with pytest.raises(ValueError, match="compress_level is not supported"):
@@ -327,6 +333,29 @@ def test_multibyte_and_non_ascii_characters_in_filenames():
     assert not zinfos[2].is_dir()
 
 
+def test_adding_windows_paths(monkeypatch):
+    """Test that paths ending with both \\ and / are considered folders on Windows"""
+
+    # mock os-specific path separators and recompute PATH_SEPARATORS to fake being on Windows
+    monkeypatch.setattr(os, "sep", ntpath.sep)
+    monkeypatch.setattr(os, "altsep", ntpath.altsep)
+    monkeypatch.setattr(zipstream.ng, "PATH_SEPARATORS", set(x for x in (os.sep, os.altsep, "/") if x))
+    zs = ZipStream(sized=True)
+    zs.add("", "a/b\\")
+    zs.add("", "a\\b/")
+    zs.add("", "a\\b\\")
+
+    predicted = len(zs)
+    actual = len(bytes(zs))
+
+    assert predicted == actual
+
+    info = zs.get_info()
+    assert len(info) == 3
+    for x in info:
+        assert x["name"] == "a/b/"
+
+
 def test_external_attrs(tmpdir):
     folder = tmpdir.mkdir("folder-frompath")
     file = tmpdir.join("file-frompath")
@@ -474,9 +503,17 @@ def test_recursion_disable(tmpdir):
     [b'a', b'list', b'of', b'bytes'],
     _gen_rand()
 ])
-def test_adding_data(data):
-    """Test adding non-files"""
-    zs = ZipStream()
+@pytest.mark.parametrize("ct", [
+    zipfile.ZIP_STORED,
+    zipfile.ZIP_LZMA,
+    zipfile.ZIP_DEFLATED,
+    zipfile.ZIP_BZIP2
+])
+def test_adding_data(caplog, data, ct):
+    """Test adding non-files with different compression methods"""
+    caplog.set_level(logging.WARNING)
+
+    zs = ZipStream(compress_type=ct)
 
     tostore = data
     if isinstance(data, str):
@@ -495,9 +532,13 @@ def test_adding_data(data):
     zs.add(tostore, "data.bin")
     zf = _get_zip(zs)
 
+    assert "Size mismatch" not in caplog.text
+
     zinfos = zf.infolist()
     assert len(zinfos) == 1
 
+    assert zinfos[0].compress_type == ct
+
     from_file = zf.read(zinfos[0])
     assert from_file == data
 
@@ -566,6 +607,68 @@ def test_adding_missing_path(tmpdir):
         ZipStream.from_path(tmpdir.join("doesntexist"))
 
 
+def test_adding_deleted_path(tmpdir):
+    zs = ZipStream(sized=True)
+    f = tmpdir.join("file")
+    f.write(b"this is some data")
+    zs.add_path(f)
+    assert len(zs) == 139
+    f.remove()
+    assert len(zs) == 139
+
+    with pytest.raises(FileNotFoundError):
+        bytes(zs)
+
+
+@pytest.mark.parametrize("data", [
+    ("short text", "then longer text"),
+    ("first longer text", "then shorter"),
+])
+@pytest.mark.parametrize("sized", (True, False))
+def test_adding_changed_path(caplog, tmpdir, sized, data):
+    caplog.set_level(logging.WARNING)
+
+    zs = ZipStream(sized=sized)
+    f = tmpdir.join("file")
+    f.write(data[0])
+    zs.add_path(f)
+    f.write(data[1])  # overwrites, doesn't append
+
+    assert not caplog.text
+    if sized:
+        with pytest.raises(RuntimeError, match="Error adding 'file' to sized ZipStream"):
+            bytes(zs)
+    else:
+        bytes(zs)
+
+    msg = "Size mismatch when adding data for 'file' (expected {} bytes, got {})".format(
+        len(data[0]),
+        len(data[1])
+    )
+    assert msg in caplog.text
+
+
+@pytest.mark.parametrize("sized", (True, False))
+@pytest.mark.parametrize("size", (None, -1, 0, 1, 100))
+def test_adding_iterator_invalid_size(caplog, sized, size):
+    """Make sure errors are raised when generating if an incorrect iterator size was provided"""
+    caplog.set_level(logging.WARNING)
+
+    l_rand = len(b''.join(_gen_rand()))
+
+    zs = ZipStream(sized=sized)
+    zs.add(_gen_rand(), "rand.txt", size=size)
+
+    assert not caplog.text
+    if sized and size is not None:
+        with pytest.raises(RuntimeError, match="Error adding 'rand.txt' to sized ZipStream"):
+            bytes(zs)
+        assert "Size mismatch when adding data for 'rand.txt'" in caplog.text
+    else:
+        bytes(zs)
+        assert "Size mismatch" not in caplog.text
+
+
 def test_adding_comment(caplog):
     caplog.set_level(logging.WARNING)
 
@@ -880,6 +983,7 @@ def test_readme_stdlib_comparison(tmpdir):
         _get_zip(ZipStream.from_path(t))
     )
 
+
 def test_add_duplicate_file():
     """Test adding multiple files with the same name works"""
     zs = ZipStream(sized=True)
@@ -900,7 +1004,7 @@ def test_add_duplicate_file():
 
 
 def test_unsized_zipstream_len_typeerror():
-    """Test that an unsized ZipStream raises a TypeError when aske for the length"""
+    """Test that an unsized ZipStream raises a TypeError when asked for the length"""
     zs = ZipStream(sized=False)
     assert not zs.sized
 
@@ -908,6 +1012,49 @@ def test_unsized_zipstream_len_typeerror():
         len(ZipStream(sized=False))
 
 
+@pytest.mark.parametrize("zip64", [False, True])
+@pytest.mark.parametrize("sized", [False, True])
+def test_proper_zip64_min_version(monkeypatch, zip64, sized):
+    """Ensure that the min version is set properly"""
+    if zip64:
+        monkeypatch.setattr(zipfile, "ZIP64_LIMIT", 100)
+        monkeypatch.setattr(zipstream.ng, "ZIP64_LIMIT", 100)
+
+    zs = ZipStream(sized=sized)
+    zs.add(_gen_rand(), "rand.txt")
+
+    zi = _get_zip(zs).infolist()[0]
+    # A sized zipstream will pull the entire iterator into memory and realize
+    # that zip64 is not required.
+    # An unsized zipstream will not pull it into memory and be forced to use
+    # zip64 extensions since the size could be too big
+    expected_version = (zipfile.DEFAULT_VERSION if sized and not zip64 else zipfile.ZIP64_VERSION)
+
+    assert zi.create_version == expected_version
+    assert zi.extract_version == expected_version
+
+
+@pytest.mark.parametrize("sized", [False, True])
+@pytest.mark.parametrize("known_size", [False, True])
+def test_adding_iterator_sized(sized, known_size):
+    """Check that adding an iterator to the ZipStream only immediately reads it
+    into memory if the ZipStream is sized and the size isn't provided
+    """
+    read = False
+    def _my_iter():
+        nonlocal read
+        yield from _gen_rand()
+        read = True
+
+    l_rand = len(b''.join(_gen_rand()))
+
+    zs = ZipStream(sized=sized)
+    zs.add(_my_iter(), "rand.txt", size=None if not known_size else l_rand)
+    assert read == (False if known_size else sized)
+    bytes(zs)
+    assert read
+
+
 @pytest.mark.parametrize("zip64", [False, True])
 def test_sized_zipstream(monkeypatch, files, zip64):
     """Test a sized ZipStream accurately calculates its final size"""
