@@ -7,7 +7,6 @@ import datetime
 import functools
 import logging
 import os
-import queue
 import stat
 import struct
 import sys
@@ -21,9 +20,9 @@ from zipfile import (
     ZIP64_VERSION, BZIP2_VERSION, ZIP_BZIP2, LZMA_VERSION, ZIP_LZMA,
     ZIP_DEFLATED,
     # Byte sequence constants
-    structCentralDir, structEndArchive64, structEndArchive,
-    structEndArchive64Locator, stringCentralDir, stringEndArchive64,
-    stringEndArchive, stringEndArchive64Locator,
+    structFileHeader, structCentralDir, structEndArchive64, structEndArchive,
+    structEndArchive64Locator, stringFileHeader, stringCentralDir,
+    stringEndArchive64, stringEndArchive, stringEndArchive64Locator,
     # Size constants
     sizeFileHeader, sizeCentralDir, sizeEndCentDir, sizeEndCentDir64Locator,
     sizeEndCentDir64,
@@ -46,6 +45,10 @@ MAX_DATE = (2107, 12, 31, 23, 59, 59)
 # be slightly bigger than uncompressed.
 ZIP64_ESTIMATE_FACTOR = 1.05
 
+# Characters that are to be considered path separators on the current platform
+# (includes "/" regardless of platform as per ZIP format specification)
+PATH_SEPARATORS = set(x for x in (os.sep, os.altsep, "/") if x)
+
 # Constants for compatibility modes
 PY36_COMPAT = sys.version_info < (3, 7)  # disable compress_level
 PY35_COMPAT = sys.version_info < (3, 6)  # backport ZipInfo functions, stringify path-like objects
@@ -86,6 +89,14 @@ def _check_compression(compress_type, compress_level):
         )
 
 
+def _timestamp_to_dos(ts):
+    """Takes an integer timestamp and converts it to a (dosdate, dostime) tuple"""
+    return (
+        (ts[0] - 1980) << 9 | ts[1] << 5 | ts[2],
+        ts[3] << 11 | ts[4] << 5 | (ts[5] // 2)
+    )
+
+
 class ZipStreamInfo(ZipInfo):
     """A ZipInfo subclass that always uses a data descriptor to store filesize data"""
 
@@ -110,13 +121,81 @@ class ZipStreamInfo(ZipInfo):
         # before it like normal. This is essential for making the zip data
         # streamable
         return struct.pack(
-            b'<4sLQQ' if zip64 else b'<4sLLL',
+            "<4sLQQ" if zip64 else "<4sLLL",
             b'PK\x07\x08',  # Data descriptor signature
             self.CRC,
             self.compress_size,
             self.file_size
         )
 
+    def FileHeader(self, zip64):
+        """Return the per-file header as bytes"""
+        # Based on code in zipfile.ZipInfo.FileHeader
+
+        # Logic for where the file sizes are listed is as follows:
+        # From the zip spec:
+        # - When using a data descriptor, the file sizes should be listed as 0
+        #   in the file header.
+        # - When using Zip64, the header size fields should always be set to
+        #   0xFFFFFFFF to indicate that the size is in the Zip64 extra field.
+        # - The format of the data descriptor depends on if a Zip64 extra field
+        #   is present in the file header.
+        # Assumption:
+        # - When using both a data descriptor and Zip64 extensions, the header
+        #   size fields should be set to 0xFFFFFFFF to indicate that the true
+        #   sizes are in the required Zip64 extra field, which should list the
+        #   sizes as 0 to defer to the data descriptor.
+
+        dosdate, dostime = _timestamp_to_dos(self.date_time)
+        if self.flag_bits & 0x08:
+            # Using a data descriptor record to record the file sizes, set
+            # everything to 0 since they'll be written there instead.
+            CRC = compress_size = file_size = 0
+        else:
+            CRC = self.CRC
+            compress_size = self.compress_size
+            file_size = self.file_size
+
+        min_version = 0
+        extra = self.extra
+        if zip64:
+            min_version = ZIP64_VERSION
+            extra += struct.pack(
+                "<HHQQ",
+                0x01,  # Zip64 extended information extra field identifier
+                16,  # length of the following "QQ" data
+                file_size,
+                compress_size,
+            )
+            # Indicate that the size is in the Zip64 extra field instead
+            file_size = 0xFFFFFFFF
+            compress_size = 0xFFFFFFFF
+
+        if self.compress_type == ZIP_BZIP2:
+            min_version = max(BZIP2_VERSION, min_version)
+        elif self.compress_type == ZIP_LZMA:
+            min_version = max(LZMA_VERSION, min_version)
+
+        self.extract_version = max(min_version, self.extract_version)
+        self.create_version = max(min_version, self.create_version)
+        filename, flag_bits = self._encodeFilenameFlags()
+        header = struct.pack(
+            structFileHeader,
+            stringFileHeader,
+            self.extract_version,
+            self.reserved,
+            flag_bits,
+            self.compress_type,
+            dostime,
+            dosdate,
+            CRC,
+            compress_size,
+            file_size,
+            len(filename),
+            len(extra)
+        )
+        return header + filename + extra
+
     def _file_data(self, iterable=None, force_zip64=False):
         """Given an iterable of file data, yield a local file header and file
         data for it.
@@ -193,9 +272,7 @@ class ZipStreamInfo(ZipInfo):
         """Yield a central directory file header for this file"""
         # Based on code in zipfile.ZipFile._write_end_record
 
-        dt = self.date_time
-        dosdate = (dt[0] - 1980) << 9 | dt[1] << 5 | dt[2]
-        dostime = dt[3] << 11 | dt[4] << 5 | (dt[5] // 2)
+        dosdate, dostime = _timestamp_to_dos(self.date_time)
         extra = []
 
         # Store sizes and offsets in the extra data if they're too big
@@ -345,7 +422,7 @@ def _iter_file(path):
             yield buf
 
 
-class ZipStream(object):
+class ZipStream:
     """A write-only zip that is generated from source files/data as it's
     iterated over.
 
@@ -380,9 +457,10 @@ class ZipStream(object):
             prior to being generated, making it work with the `len()` function.
             Enabling this will enforce two restrictions:
               - No compression can be used
-              - Any iterables added to the stream will immediately be read fully
-                into memory since the size of them needs to be known at the time
-                they are added.
+              - Any iterables added to the stream without also specifying their
+                size (see `.add` docs) will immediately be read fully into
+                memory. This is because the size of the data they will produce
+                must be known prior to the stream being generated.
 
             If `False` (the default), no restrictions are enforced and using the
             object with the `len()` function will not work (will raise a
@@ -571,7 +649,7 @@ class ZipStream(object):
 
     @_validate_final
     @_validate_compression
-    def add(self, data, arcname, *, compress_type=None, compress_level=None):
+    def add(self, data, arcname, *, size=None, compress_type=None, compress_level=None):
         """Queue up data to be added to the ZipStream
 
         `data` can be bytes, a string (encoded to bytes using utf-8), or any
@@ -585,6 +663,10 @@ class ZipStream(object):
         any `data` is provided then the `arcname` cannot end with a "/" as this
         would create a directory (which can't contain content).
 
+        `size` (optional) specifies the size of the `data` ONLY in the case
+        where it is an iterator and the ZipStream is sized. It is ignored in
+        all other cases.
+
         Note that the data provided will not be used until the file is actually
         encoded in the ZipStream. This means that strings and bytes will be held
         in memory and iterables will not be iterated over until then. For this
@@ -610,8 +692,7 @@ class ZipStream(object):
             # change while we're iterating over it.
             data = bytes(data)
 
-        # Check platform-specific separator as well as zip-specific
-        if data != b'' and arcname[-1] in (os.sep, "/"):
+        if data != b'' and arcname[-1] in PATH_SEPARATORS:
             raise ValueError("Can't store data as a directory")
 
         if isinstance(data, bytes):
@@ -624,6 +705,7 @@ class ZipStream(object):
         elif hasattr(data, "__iter__"):
             self._enqueue(
                 iterable=data,
+                size=size if self._sized else None,
                 arcname=arcname,
                 compress_type=compress_type,
                 compress_level=compress_level,
@@ -645,7 +727,7 @@ class ZipStream(object):
         Returns a `datetime.datetime` object or `None` if the ZipStream is
         empty.
         """
-        return datetime.datetime(*self._last_modified) if self else None
+        return datetime.datetime(*self._last_modified) if self._last_modified else None
 
     @property
     def comment(self):
@@ -742,6 +824,8 @@ class ZipStream(object):
         """Internal method to enqueue files, data, and iterables to be streamed"""
 
         path = kwargs.get("path")
+        data = kwargs.get("data")
+        size = kwargs.get("size")
 
         # Get the modified time of the added path (use current time for
         # non-paths) and use it to update the last_modified property
@@ -749,29 +833,31 @@ class ZipStream(object):
         if self._last_modified is None or self._last_modified < mtime:
             self._last_modified = mtime
 
+        # Get the expected size of the data where not specified and possible
+        if size is None:
+            if data is not None:
+                kwargs["size"] = len(data)
+            elif path is not None:
+                if os.path.isdir(path):
+                    kwargs["size"] = 0
+                else:
+                    kwargs["size"] = os.path.getsize(path)
+
         # If the ZipStream is sized then it will look at what is being added and
         # queue up some information for _get_size to use to compute the total
         # length of the stream. It will also read any iterables fully into
         # memory so their size is known.
         if self._sized:
-            iterable = kwargs.pop("iterable", None)
-            data = kwargs.get("data")
-
-            assert bool(iterable) ^ bool(data is not None) ^ bool(path)
-
             if kwargs.get("compress_type"):
                 raise ValueError("Cannot use compression with a sized ZipStream")
-            elif data is not None:
-                filesize = len(data)
-            elif iterable:
-                # Iterate the data to get the size and replace it with the static data
-                data = b''.join(iterable)
-                filesize = len(data)
-                kwargs["data"] = data
-            elif path:
-                filesize = os.path.getsize(path)
 
-            self._to_count.append((kwargs["arcname"], filesize))
+            # Iterate the iterable data to get the size and replace it with the static data
+            if path is None and data is None and size is None:
+                data = b''.join(kwargs.pop("iterable"))
+                kwargs["size"] = len(data)
+                kwargs["data"] = data
+
+            self._to_count.append((kwargs["arcname"], kwargs["size"]))
 
         self._queue.append(kwargs)
 
@@ -780,9 +866,10 @@ class ZipStream(object):
         self._pos += len(data)
         return data
 
-    def _gen_file_entry(self, *, path=None, iterable=None, data=None, arcname, compress_type, compress_level):
+    def _gen_file_entry(self, *, path=None, iterable=None, data=None, size=None, arcname, compress_type, compress_level):
         """Yield the zipped data generated by the specified path/iterator/data"""
         assert bool(path) ^ bool(iterable) ^ bool(data is not None)
+        assert not (self._sized and size is None)
 
         if path:
             zinfo = ZipStreamInfo.from_file(path, arcname)
@@ -823,6 +910,25 @@ class ZipStream(object):
         # Generate the file data
         for x in zinfo._file_data(iterable, force_zip64=force_zip64):
             yield self._track(x)
+
+        if size is not None and size != zinfo.file_size:
+            # The size of the data that was stored didn't match what was
+            # expected. Note that this still produces a valid zip file, just
+            # one with a different amount of data than was expected.
+            # If the ZipStream is sized, this will raise an error since the
+            # actual size will no longer match the calculated size.
+            __log__.warning(
+                "Size mismatch when adding data for '%s' (expected %d bytes, got %d)",
+                arcname,
+                size,
+                zinfo.file_size
+            )
+            if self._sized:
+                raise RuntimeError(
+                    "Error adding '{}' to sized ZipStream - "
+                    "actual size did not match the computed size".format(arcname)
+                )
+
         self._filelist.append(zinfo)
 
     def _gen_archive_footer(self):
@@ -928,7 +1034,7 @@ class ZipStream(object):
                 files_size += sizeFileHeader + arcname_len # 30 + name len
 
                 # Folders don't have any data or require any extra records
-                if arcname[-1] not in (os.sep, "/"):
+                if arcname[-1] not in PATH_SEPARATORS:
 
                     # When using zip64, the size and compressed size of the file are
                     # written as an extra field in the FileHeader.
